import * as ini from "ini";
import * as fs from "fs";

type IGenericField<T, Co> =
   | Co
   | {
        type: Co;
        default?: T;
        optional?: boolean;
        description?: string;
     };

type IValidTypes = string | number | boolean;

type ITestField<
   T1 extends IValidTypes,
   T2 extends IValidTypes,
   R
> = T1 extends T2 ? R : never;

export type IField<F extends IValidTypes> =
   | ITestField<F, string, IGenericField<string, StringConstructor>>
   | ITestField<F, number, IGenericField<number, NumberConstructor>>
   | ITestField<F, boolean, IGenericField<boolean, BooleanConstructor>>;

export type IConfigInterface = {
   [category: string]: {
      [field: string]: IValidTypes | undefined;
   };
};
export const Optional = Symbol("optional");

export type IOptional<T extends IConfigInterface> = {
   [category in keyof T]: {
      [Optional]?: boolean;
   };
};

export type IFields<T extends IConfigInterface> = {
   [category in keyof T]: {
      [field in keyof T[category]]: IField<T[category][field]>;
   };
};

export type IConfig<T extends IConfigInterface> = IFields<T> | IOptional<T>;
function parseData(config: IConfig<any>, rawdata: string) {
   const data = rawdata ? ini.parse(rawdata) : {};
   let result: any = {};
   let errors: string[] = [];
   for (let C in config) {
      if (config[C][<any>Optional]) {
         if (
            !data[C] &&
            !Object.keys(process.env).find((key) =>
               key.startsWith(`${C.toUpperCase}_`)
            )
         ) {
            continue;
         }
      }

      result[C] = {};
      for (let F in config[C]) {
         let gfieldConfig = config[C][F];
         let fieldConfig: Exclude<
            typeof gfieldConfig,
            NumberConstructor | StringConstructor | BooleanConstructor
         > = (gfieldConfig === Number ||
         gfieldConfig === String ||
         gfieldConfig === Boolean
            ? { type: gfieldConfig }
            : gfieldConfig) as any;

         const envname = `${C.toUpperCase()}_${F.toUpperCase()}`;
         let value =
            (data[C] || {})[F] || process.env[envname] || fieldConfig.default;

         if (value !== undefined && value !== null) {
            switch (fieldConfig.type) {
               case Number:
                  value = Number(value);
                  break;
               case String:
                  value = String(value);
                  break;
               case Boolean:
                  value = value === "true";
                  break;
            }
         } else if (!fieldConfig.optional) {
            errors.push(`[${C}][${F}] Cannot be undefined!`);
         }

         result[C][F] = value;
      }
   }

   if (errors.length > 0)
      throw new Error("There are errors: \n" + errors.join("\n"));

   return result;
}

function parse<T extends IConfigInterface>(config: IConfig<T>, path?: string) {
   let data;
   if (path)
      data = fs.existsSync(path)
         ? fs.readFileSync(path).toString("utf8")
         : undefined;
   return parseData(config, data) as T;
}

async function parseAsync<T extends IConfigInterface>(
   config: IConfig<T>,
   path: string
) {
   let data;
   if (path && (await new Promise((yes) => fs.exists(path, yes)))) {
      data = await new Promise<Buffer>((yes, no) =>
         fs.readFile(path, (err, data) => (err ? no(err) : yes(data)))
      ).then((res) => res.toString("utf8"));
   }
   return parseData(config, data) as T;
}

const typeNameMap = new Map<any, string>();
typeNameMap.set(String, "string");
typeNameMap.set(Number, "number");
typeNameMap.set(Boolean, "boolean");

function getOptionsHelp(config: IConfig<any>) {
   let ini: string[] = [];
   let docker: string[] = [];
   for (const category in config) {
      ini.push(`[${category}]`);
      for (const key in config[category]) {
         let entry = config[category][key];
         let field: Exclude<
            typeof entry,
            NumberConstructor | StringConstructor | BooleanConstructor
         > = (entry === Number || entry === String || entry === Boolean
            ? { type: entry }
            : entry) as any;

         let typeName = typeNameMap.get(field.type);
         let optional = !!field.optional;
         let description = field.description || "";

         let def = field.default ? ` (default: ${field.default})` : "";

         ini.push(
            `${key}=<${optional ? "?" : ""}${typeName}>${def} ${description}`
         );
         docker.push(
            `${category.toUpperCase()}_${key.toUpperCase()}=<${
               optional ? "?" : ""
            }${typeName}>${def} ${description}`
         );
      }
   }

   return { ini: ini.join("\n"), docker: docker.join("\n") };
}

function printHelp(config: IConfig<any>) {
   const help = getOptionsHelp(config);

   console.log("ini config:");
   console.log(help.ini);

   console.log("");
   console.log("Docker config:");
   console.log(help.docker);
}

export { parse, parseAsync, getOptionsHelp, printHelp };
